import asyncio


def async_test(f):
    def wrapper(*args, **kwargs):
        future = f(*args, **kwargs)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(future)

    return wrapper



import unittest

from web3 import Web3
from web3.types import TxReceipt

from jasmine_eth import SDK
from jasmine_eth.sdk import Account, _Web3Wrapper
from tests.utils import async_test


class AccountTests(unittest.TestCase):
    def setUp(self) -> None:
        self.web3 = Web3(Web3.HTTPProvider("http://localhost:8545"))

    def test_instantiate_account(self):
        private_key = "0x787ff27e287db81a926333f85bfedd67c497453a08b278afcea0fe9f4c0e235d"
        account = Account(self.web3, private_key)
        self.assertEqual(account.private_key, private_key)
        self.assertEqual(account.address, "0x29B2F1587BCF319eBB510284aFd552F995410C8a")


class Web3WrapperTests(unittest.TestCase):
    def setUp(self) -> None:
        self.web3 = Web3(Web3.HTTPProvider("http://localhost:8545"))
        self.wrapper = _Web3Wrapper(self.web3)

    @async_test
    async def test_send_transaction(self):
        key = "0x4f3edf983ac636a65a842ce7c78d9aa706d3b113bce9c46f30d7d21715b23b1d"
        account = Account(self.web3, key)
        receipt: TxReceipt = await self.wrapper.send_transaction({
            "from": account.address,
            "to": account.address,
            "value": 0,
        }, account)
        self.assertEqual(receipt["from"], account.address)


class SDKTests(unittest.TestCase):
    def setUp(self) -> None:
        self.sdk = SDK("http://localhost:8545")
        self.accounts = [Account(self.sdk.web3, private_key) for private_key in [
            "0x4f3edf983ac636a65a842ce7c78d9aa706d3b113bce9c46f30d7d21715b23b1d",
            "0x6cbed15c793ce57650b9877cf6fa156fbef513c4e6134f022a85b1ffdd59b2a1",
            "0x6370fd033278c143179d81c5526140625662b8daa446c22ee2d73db3707e620c",
        ]]

    def test_retrieve_account(self):
        key = "0x4f3edf983ac636a65a842ce7c78d9aa706d3b113bce9c46f30d7d21715b23b1d"
        account = self.sdk.retrieve_account(key)
        self.assertEqual(account.private_key, key)
        self.assertEqual(account.address, "0x90F8bf6A479f320ead074411a4B0e7944Ea8c9C1")

    def test_create_account(self):
        account = self.sdk.create_account()
        self.assertEqual(account.address, self.sdk.web3.eth.account.from_key(account.private_key).address)

    def test_balance_of(self):
        new_account = self.sdk.create_account()
        balance = self.sdk.balance_of(new_account.address)
        self.assertEqual(balance, 0)
        balance = self.sdk.balance_of(self.accounts[2].address)
        self.assertEqual(balance, 100000000000000000000)

    def test_eth_to_wei(self):
        wei = self.sdk.eth_to_wei(1)
        self.assertEqual(wei, 1000000000000000000)

    def test_wei_to_eth(self):
        eth = self.sdk.wei_to_eth(1000000000000000000)
        self.assertEqual(eth, 1)

    @async_test
    async def test_transfer(self):
        balance_before = self.sdk.balance_of(self.accounts[1].address)
        await self.sdk.transfer(self.accounts[1].address, self.sdk.eth_to_wei(1), self.accounts[0])
        balance_after = self.sdk.balance_of(self.accounts[1].address)
        self.assertEqual(balance_after - balance_before, self.sdk.eth_to_wei(1))

    @async_test
    async def test_deploy_tfc_manager(self):
        manager_address = await self.sdk.deploy_tfc_manager(self.accounts[0])
        code = self.sdk.web3.eth.getCode(manager_address)
        self.assertTrue(len(code.hex()) > 2)


class TFCManagerTests(unittest.TestCase):
    @classmethod
    @async_test
    async def setUp(cls) -> None:
        cls.sdk = SDK("http://localhost:8545")
        cls.accounts = [Account(cls.sdk.web3, private_key) for private_key in [
            "0x4f3edf983ac636a65a842ce7c78d9aa706d3b113bce9c46f30d7d21715b23b1d",
            "0x6cbed15c793ce57650b9877cf6fa156fbef513c4e6134f022a85b1ffdd59b2a1",
            "0x6370fd033278c143179d81c5526140625662b8daa446c22ee2d73db3707e620c",
        ]]
        manager_address = await cls.sdk.deploy_tfc_manager(cls.accounts[0])
        cls.manager = cls.sdk.get_tfc_manager(manager_address)
        tfc_address = cls.manager.tfc_token_address()
        cls.tfc = cls.sdk.get_tfc_token(tfc_address)

    def test_tfc_token_address(self):
        address = self.manager.tfc_token_address()
        code = self.sdk.web3.eth.getCode(address)
        self.assertTrue(len(code.hex()) > 2)

    @async_test
    async def test_claim_tfc(self):
        # invalid signature
        sig = "0x123456"
        try:
            await self.manager.claim_tfc(1, 0, sig, self.accounts[0])
            self.assertFalse(True)
        except Exception as e:
            self.assertIsNotNone(e)

        # wrong claimer
        sig = "0x6b04573d9a5b813e65b7afc77ca931bf1a5787ed1732622034355c75b39fe934194501f3431b2fed46581eaa486cdb636eebcb7f852d2105af4a4b53a25dd27e1c"
        try:
            await self.manager.claim_tfc(1, 0, sig, self.accounts[0])
            self.assertFalse(True)
        except Exception as e:
            self.assertIsNotNone(e)

        # successful
        sig = "0x6b04573d9a5b813e65b7afc77ca931bf1a5787ed1732622034355c75b39fe934194501f3431b2fed46581eaa486cdb636eebcb7f852d2105af4a4b53a25dd27e1c"
        try:
            balance = self.tfc.balance_of(self.accounts[1].address)
            self.assertEqual(balance, 0)
            await self.manager.claim_tfc(1, 0, sig, self.accounts[1])
            balance = self.tfc.balance_of(self.accounts[1].address)
            self.assertEqual(balance, 1)
        except Exception as e:
            self.assertIsNone(e)

        # used nonce
        sig = "0x6b04573d9a5b813e65b7afc77ca931bf1a5787ed1732622034355c75b39fe934194501f3431b2fed46581eaa486cdb636eebcb7f852d2105af4a4b53a25dd27e1c"
        try:
            await self.manager.claim_tfc(1, 0, sig, self.accounts[1])
        except Exception as e:
            self.assertIsNotNone(e)


class TFCTokenTests(unittest.TestCase):
    @classmethod
    @async_test
    async def setUpClass(cls) -> None:
        cls.sdk = SDK("http://localhost:8545")
        cls.accounts = [Account(cls.sdk.web3, private_key) for private_key in [
            "0x4f3edf983ac636a65a842ce7c78d9aa706d3b113bce9c46f30d7d21715b23b1d",
            "0x6cbed15c793ce57650b9877cf6fa156fbef513c4e6134f022a85b1ffdd59b2a1",
            "0x6370fd033278c143179d81c5526140625662b8daa446c22ee2d73db3707e620c",
        ]]
        manager_address = await cls.sdk.deploy_tfc_manager(cls.accounts[0])
        cls.manager = cls.sdk.get_tfc_manager(manager_address)
        tfc_address = cls.manager.tfc_token_address()
        cls.tfc = cls.sdk.get_tfc_token(tfc_address)
        # faucet 1e-18 TFC
        sig = "0x6b04573d9a5b813e65b7afc77ca931bf1a5787ed1732622034355c75b39fe934194501f3431b2fed46581eaa486cdb636eebcb7f852d2105af4a4b53a25dd27e1c"
        await cls.manager.claim_tfc(1, 0, sig, cls.accounts[1])
        balance = cls.tfc.balance_of(cls.accounts[1].address)
        assert balance == 1

    def test_properties(self):
        self.assertEqual(self.tfc.name, "TFCToken")
        self.assertEqual(self.tfc.symbol, "TFC")
        self.assertEqual(self.tfc.total_supply, 1)
        self.assertEqual(self.tfc.decimals, 18)

    def test_balance_of(self):
        self.assertEqual(self.tfc.balance_of(self.accounts[0].address), 0)
        self.assertEqual(self.tfc.balance_of(self.accounts[1].address), 1)

    def test_allowance(self):
        self.assertEqual(self.tfc.allowance(self.accounts[0].address, self.accounts[1].address), 0)

    @async_test
    async def test_transfer(self):
        # transfer from accounts[1] to accounts[0]
        try:
            await self.tfc.transfer(self.accounts[0].address, 1, self.accounts[1])
            balance = self.tfc.balance_of(self.accounts[0].address)
            self.assertEqual(balance, 1)
            balance = self.tfc.balance_of(self.accounts[1].address)
            self.assertEqual(balance, 0)
        except Exception as e:
            self.assertIsNone(e)

        # accounts[1] approve accounts[1]
        try:
            await self.tfc.approve(self.accounts[1].address, 1, self.accounts[0])
            allowance = self.tfc.allowance(self.accounts[0].address, self.accounts[1].address)
            self.assertEqual(allowance, 1)
        except Exception as e:
            self.assertIsNone(e)

        # transfer from accounts[0] to accounts[1]
        try:
            await self.tfc.transfer_from(self.accounts[0].address, self.accounts[1].address, 1, self.accounts[1])
            balance = self.tfc.balance_of(self.accounts[0].address)
            self.assertEqual(balance, 0)
            balance = self.tfc.balance_of(self.accounts[1].address)
            self.assertEqual(balance, 1)
        except Exception as e:
            self.assertIsNone(e)



import asyncio
import json
from os import path

from web3.gas_strategies.rpc import rpc_gas_price_strategy
from web3.types import TxParams, TxReceipt, ABI
from web3 import Web3


class _Web3Wrapper(object):
    def __init__(self, web3: Web3):
        self._web3 = web3

    @property
    def web3(self) -> Web3:
        return self._web3

    async def send_transaction(self, transaction: TxParams, sender) -> TxReceipt:
        # TODO implement transaction confirmation requirement
        loop = asyncio.get_event_loop()
        future = loop.create_future()

        # check transaction fields
        if "gas" not in transaction:
            transaction["gas"] = self._web3.eth.estimateGas(transaction)
        if "gasPrice" not in transaction:
            self._web3.eth.setGasPriceStrategy(rpc_gas_price_strategy)
            transaction["gasPrice"] = self._web3.eth.generateGasPrice(transaction)
        if "nonce" not in transaction:
            transaction["nonce"] = self._web3.eth.getTransactionCount(sender.address)

        # sign transaction
        signed_tx = self._web3.eth.account.sign_transaction(transaction, sender.private_key)

        async def transaction_task():
            try:
                tx_hash = self._web3.eth.sendRawTransaction(signed_tx.rawTransaction)
                receipt: TxReceipt = self._web3.eth.waitForTransactionReceipt(tx_hash)
                future.set_result(receipt)
            except Exception as e:
                future.set_exception(e)

        loop.create_task(transaction_task())
        return await future


class Account(_Web3Wrapper):
    """
    Represent an Ethereum account.
    """

    def __init__(self, web3: Web3, private_key: str):
        """
        :param web3: web3py web3 instance
        :param private_key: the private key of the account
        """
        super().__init__(web3)
        self._web3 = web3
        self._private_key = private_key
        self._eth_account = self._web3.eth.account.from_key(private_key)

    @property
    def private_key(self) -> str:
        """
        :return: the private key (str) of the account
        """
        return self._private_key

    @property
    def address(self) -> str:
        """
        :return: the Ethereum address (str) of the account
        """
        return self._eth_account.address


class SDK(_Web3Wrapper):
    """
    The jasmine-eth SDK class
    """

    def __init__(self, endpoint: str):
        """
        :param endpoint: Ethereum endpoint. e.g. http://localhost:8545
        """
        # initiate web3
        endpoint = endpoint.strip()
        if endpoint is not None and endpoint.startswith("http"):
            provider = Web3.HTTPProvider(endpoint)
        elif endpoint is not None and endpoint.startswith("ws"):
            provider = Web3.WebsocketProvider(endpoint)
        else:
            raise ValueError("unsupported Ethereum endpoint {}".format(endpoint))
        super().__init__(Web3(provider))

    def create_account(self) -> Account:
        """
        Create a new account
        :return: a new Ethereum account
        """
        acc = self.web3.eth.account.create()
        return Account(self.web3, acc.key)

    def retrieve_account(self, private_key: str) -> Account:
        """
        Retrieve an existing account using private key
        :return: the retrieved Ethereum account
        """
        return Account(self.web3, private_key)

    def balance_of(self, address: str) -> int:
        """
        Get the ETH balance of an account
        :param address: the address of the account
        :return: ETH balance in wei
        """
        return self.web3.eth.getBalance(address)

    async def transfer(self, recipient: str, amount: int, sender: Account):
        """
        Transfer the amount of ETH from sender to recipient.

        This is an async function will complete only when the transaction is executed on chain.
        If the transaction fails, exception will be thrown.
        :param recipient: the address of recipient address
        :param amount: the amount of ETH to transfer (in wei)
        :param sender: the sender account
        """
        transaction: TxParams = {
            "from": sender.address,
            "to": recipient,
            "value": amount,
        }
        await self.send_transaction(transaction, sender)

    def wei_to_eth(self, amount_wei: int) -> float:
        """
        Convert from wei to ETH
        :param amount_wei: amount of wei
        :return: the amount of ETH according to the amount of wei
        """
        return self.web3.fromWei(amount_wei, 'ether')

    def eth_to_wei(self, amount_eth: float) -> int:
        """
        Convert from ETH to wei
        :param amount_eth: amount of eth
        :return: the amount of wei according to the amount of ETH
        """
        return self.web3.toWei(amount_eth, 'ether')

    async def deploy_tfc_manager(self, deployer: Account) -> str:
        """
        Deploy the TFCManager smart contract, which will implicitly deploy a TFCToken contract.

        This is an async function will complete only when the deployment is executed on chain.
        If the deployment fails, exception will be thrown.
        :param deployer: the Ethereum account used to deploy
        :return: the address of TFCManager contract
        """
        contract = self.web3.eth.contract(abi=TFCManager.abi(), bytecode=TFCManager.bytecode())
        transaction = contract.constructor().buildTransaction({
            "from": deployer.address,
        })
        receipt: TxReceipt = await self.send_transaction(transaction, deployer)
        return receipt["contractAddress"]

    def get_tfc_manager(self, address: str):
        """
        Get the TFCManager contract instance
        :param address: the address of TFCManager contract
        :return: TFCManager contract instance
        """
        return TFCManager(self.web3, address)

    def get_tfc_token(self, address: str):
        """
        Get the TFCToken contract instance
        :param address: the address of TFCToken contract
        :return: TFCToken contract instance
        """
        return TFCToken(self.web3, address)


class TFCManager(_Web3Wrapper):
    @staticmethod
    def bytecode() -> str:
        dir_path = path.dirname(path.realpath(__file__))
        with open(path.join(dir_path, "contracts", "TFCManager.bin")) as file:
            return file.readline()

    @staticmethod
    def abi() -> ABI:
        dir_path = path.dirname(path.realpath(__file__))
        with open(path.join(dir_path, "contracts", "TFCManager.abi.json")) as file:
            return json.load(file)

    def __init__(self, web3: Web3, address: str):
        super().__init__(web3)
        self._contract = self.web3.eth.contract(address=address, abi=self.abi())

    def tfc_token_address(self) -> str:
        """
        Get the TFCToken contract address linked to this TFCManager
        :return: address of TFCToken contract
        """
        return self._contract.functions.tfcToken().call()

    async def claim_tfc(self, amount: int, nonce: int, signature: str, claimer: Account):
        """
        Claim TFC ERC20 token.
        It takes the amount of TFC token to claim, the nonce used and the signature signed by TFCManager admin account.
        The amount of TFC token is in the minimal unit 10^-18.

        This is an async function will complete only when the claim transaction is executed on chain.
        If the claim transaction fails, exception will be thrown.
        :param amount: the amount of TFC token to claim (unit 10^-18)
        :param nonce: the nonce used in this claim
        :param signature: the signature signed by TFCManager admin account
        :param claimer: the Ethereum account which wants to claim these TFC
        """
        sig = self.web3.toBytes(hexstr=signature)
        transaction = self._contract.functions.claimTFC(amount, nonce, sig).buildTransaction({
            "from": claimer.address
        })
        await self.send_transaction(transaction, claimer)


class TFCToken(_Web3Wrapper):
    @staticmethod
    def bytecode() -> str:
        dir_path = path.dirname(path.realpath(__file__))
        with open(path.join(dir_path, "contracts", "TFCToken.bin")) as file:
            return file.readline()

    @staticmethod
    def abi() -> ABI:
        dir_path = path.dirname(path.realpath(__file__))
        with open(path.join(dir_path, "contracts", "TFCToken.abi.json")) as file:
            return json.load(file)

    def __init__(self, web3: Web3, address: str):
        super().__init__(web3)
        self._contract = self.web3.eth.contract(address=address, abi=self.abi())

    @property
    def name(self):
        """
        :return: the name of this ERC20 contract
        """
        return self._contract.functions.name().call()

    @property
    def symbol(self):
        """
        :return: the symbol of this ERC20
        """
        return self._contract.functions.symbol().call()

    @property
    def decimals(self):
        """
        :return: the decimals used by this ERC20
        """
        return self._contract.functions.decimals().call()

    @property
    def total_supply(self):
        """
        :return: total supply of this ERC20
        """
        return self._contract.functions.totalSupply().call()

    def allowance(self, owner: str, spender: str) -> int:
        """
        Get the allowance of TFC token which the owner allows the spender to spend
        :param owner: the owner address
        :param spender: the spender address
        :return: the amount of allowance (in unit 10^-18)
        """
        return self._contract.functions.allowance(owner, spender).call()

    def balance_of(self, owner: str) -> int:
        """
        Get the TFC token balance of an account
        :param owner: the address of the account
        :return: the amount of TFC balance (in unit 10^-18)
        """
        return self._contract.functions.balanceOf(owner).call()

    async def transfer(self, recipient: str, amount: int, sender: Account):
        """
        Transfer the amount of TFC from sender to recipient.

        This is an async function will complete only when the transaction is executed on chain.
        If the transaction fails, exception will be thrown.
        :param recipient: the address of recipient
        :param amount: the amount of TFC to transfer (in unit 10^-18)
        :param sender: the sender account
        """
        transaction = self._contract.functions.transfer(recipient, amount).buildTransaction({
            "from": sender.address
        })
        await self.send_transaction(transaction, sender)

    async def transfer_from(self, sender: str, recipient: str, amount: int, spender: Account):
        """
        Spender transfer the amount of TFC from sender to recipient.

        This is an async function will complete only when the transaction is executed on chain.
        If the transaction fails, exception will be thrown.
        :param sender: the sender address
        :param recipient: the address of recipient
        :param amount: the amount of TFC to transfer (in unit 10^-18)
        :param spender: the account which has the allowance to transfer TFC from the sender
        """
        transaction = self._contract.functions.transferFrom(sender, recipient, amount).buildTransaction({
            "from": spender.address
        })
        await self.send_transaction(transaction, spender)

    async def approve(self, spender: str, amount: int, owner: Account):
        """
        Approve the amount of TFC for spender to spend. (Give the allowance)
        :param spender: the spender address
        :param amount: the amount of TFC to approve (in unit 10^-18)
        :param owner: the owner account
        :return:
        """
        transaction = self._contract.functions.approve(spender, amount).buildTransaction({
            "from": owner.address
        })
        await self.send_transaction(transaction, owner)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="jasmine_eth",  # Replace with your own username
    version="0.0.3",
    author="Troublor",
    author_email="troublor@live.com",
    description="Jasmine Project Ethereum SDK (Python)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Troublor/jasmine-eth-python",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'web3==5.13.1'
    ]
)
